(*  Copyright (C) 2011  Vincent Gripon (name.surname@ens-cachan.org)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.*)

(** Very small implementation for binary arrays (or "bitarrays"). These require a lot less memory than boolean arrays (the latter use 32 or 64 bits representation for each value depending on the system). *)
open Bigarray

(** Internal type for the representation of binary arrays. *)
type t = 
    {
      array:(int, Bigarray.int16_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t;
      length:int;
    }

(** Return the length of a binary array. *)
let length (bitarray:t) =
  bitarray.length

(** Used for internal computation. *)
let values,size = 
  [|1;2;4;8;16;32;64;128;256;512;1024;2048;4096;8192;16384;32768|],16

(** Create a fresh binary array of given length and with given boolean value. *)
let make length default_value =
  if length < 0
  then raise (Invalid_argument "Bitarray.make");
  let virtual_length = length / size + if length mod size != 0 then 1 else 0 in
  let array = Array1.create int16_unsigned c_layout virtual_length in
  for i = 0 to virtual_length - 1 do
    array.{i} <- if default_value then 65535 else 0;
  done;
  {
    array = array;
    length = length
  }

(** Set the indexed cell in a binary array to the specified value. *)
let set bitarray index value =
  if index < 0 || index >= bitarray.length
  then raise (Invalid_argument "Bitarray.get");
  let array = bitarray.array in
  let virtual_index = index / size in
  let shift_index = index mod size in
  array.{virtual_index} <- if value then array.{virtual_index} lor values.(shift_index) else array.{virtual_index} land (lnot values.(shift_index))

(** Get the value of the indexed cell in a binary array. *)
let get bitarray index =
  if index < 0 || index >= bitarray.length
  then raise (Invalid_argument "Bitarray.get");
  let array = bitarray.array in
  let virtual_index = index / size in
  let shift_index = index mod size in
  array.{virtual_index} land values.(shift_index) != 0

(** Iterator with index for binary arrays. *)
let iteri f bitarray =
  let array = bitarray.array in
  let vlength = bitarray.length / size + if bitarray.length mod size = 0 then 0 else 1 in
  for i = 0 to vlength - 1 do
    let delta = i * size in
    if i = vlength - 1 
    then 
      let modulo = (length bitarray) mod size in
      Array.iteri (fun index x -> if index <= modulo || modulo = 0 then f (delta + index) ((x land array.{i})!=0)) values 
    else 
      Array.iteri (fun index x -> f (delta+index) ((x land array.{i}) != 0)) values; 
  done

(** Iterator for binary arrays. *)
let iter f bitarray =
  iteri (fun _ x -> f x) bitarray

(** Create a fresh binary array of given length and using given indexed function to fill cells. *)
let init length f =
  let bitarray = make length false in
  for i = 0 to length - 1 do
    set bitarray i (f i)
  done;
  bitarray

(** Append two binary arrays to a fresh one. *)
let append bitarraya bitarrayb =
  let new_length = bitarraya.length + bitarrayb.length in
  let bitarray = make new_length false in
  for i = 0 to bitarraya.length - 1 do
    set bitarray i (get bitarraya i)
  done;
  for i = 0 to bitarrayb.length - 1 do
    set bitarray (i + bitarraya.length) (get bitarrayb i)
  done;
  bitarray

(** Concat a list of binary arrays. *)
let concat bitarray_list = 
  List.fold_left (fun a b -> append a b) (make 0 false) bitarray_list

(** Return a sub array from starting index and with given length. *)
let sub bitarray start length =
  let res = make length false in
  for i = 0 to length - 1 do
    set res i (get bitarray (i + start))
  done;
  res

(** Return a fresh copy of a binary array. *)
let copy bitarray =
  let res = make bitarray.length false in
  for i = 0 to bitarray.length - 1 do
    set res i (get bitarray i)
  done;
  res

(** Fill bitarray from given index and for given length with specified value. *)
let fill bitarray offset length value =
  for i = offset to offset + length - 1 do
    set bitarray i value
  done

